from colorsys import hsv_to_rgb as _hsv_to_rgb
from colorsys import rgb_to_hsv as _rgb_to_hsv
from typing import Tuple, Union, Optional

from typing_extensions import Self


class ColorType:
    pass


class RGB(ColorType):
    _RGB_MAX: int = 2 ** 8 - 1

    def __init__(self, r: int, g: int, b: int):
        max_value = self._RGB_MAX
        if not 0 <= r <= max_value or not 0 <= g <= max_value or not 0 <= b <= max_value:
            raise ValueError()

        self._r: int = r
        self._g: int = g
        self._b: int = b

    @property
    def red(self) -> int:
        """red value"""
        return self._r

    @property
    def r(self) -> int:
        """red value"""
        return self._r

    @property
    def green(self) -> int:
        """green value"""
        return self._g

    @property
    def g(self) -> int:
        """green value"""
        return self._g

    @property
    def blue(self) -> int:
        """blue value"""
        return self._b

    @property
    def b(self) -> int:
        """blue value"""
        return self._b

    def replace(self, r: Optional[int] = None, g: Optional[int] = None, b: Optional[int] = None,
                red: Optional[int] = None, green: Optional[int] = None, blue: Optional[int] = None) -> Self:
        """Create a new object with (optionally) replaced values.

        :param r: new red value
        :param red: new red value
        :param g: new green value
        :param green: new green value
        :param b: new blue value
        :param blue: new blue value
        """

        if r is not None:
            if red is not None:
                raise ValueError()
            red = r

        if g is not None:
            if green is not None:
                raise ValueError()
            green = g

        if b is not None:
            if blue is not None:
                raise ValueError()
            blue = b

        return self.__class__(
            red if red is not None else self._r,
            green if green is not None else self._g,
            blue if blue is not None else self._b,
        )

    def __str__(self):
        return f'{self.__class__.__name__}({self._r:d}, {self._g:d}, {self._b})'

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._r == other._r and self._g == other._g and self._b == other._b
        elif isinstance(other, HSB):
            return self == self.__class__.from_hsb(other)
        else:
            return NotImplemented

    def __getitem__(self, item: int) -> int:
        if item == 0:
            return self._r
        if item == 1:
            return self._g
        if item == 2:
            return self._b
        raise IndexError()

    # ------------------------------------------------------------------------------------------------------------------
    # Conversions
    # ------------------------------------------------------------------------------------------------------------------
    def to_hsb(self) -> 'HSB':
        """Create a new HSB object from this object

        :return: New HSB object
        """
        max_value = self._RGB_MAX
        h, s, v = _rgb_to_hsv(self._r / max_value, self._g / max_value, self._b / max_value)
        return HSB(h * HUE_FACTOR, s * PERCENT_FACTOR, v * PERCENT_FACTOR)

    @classmethod
    def from_hsb(cls, obj: Union['HSB', Tuple[float, float, float]]) -> Self:
        """Return new Object from a HSB object for a hsb tuple

        :param obj: HSB object or tuple with HSB values
        :return: new RGB object
        """
        max_value = cls._RGB_MAX
        h, s, b = (obj._hue, obj._saturation, obj._brightness) if isinstance(obj, HSB) else obj
        r, g, b = _hsv_to_rgb(h / HUE_FACTOR, s / PERCENT_FACTOR, b / PERCENT_FACTOR)
        return cls(round(r * max_value), round(g * max_value), round(b * max_value))


class RGB16(RGB):
    _RGB_MAX: int = 2 ** 16 - 1


class RGB24(RGB):
    _RGB_MAX: int = 2 ** 32 - 1


class RGB32(RGB):
    _RGB_MAX: int = 2 ** 32 - 1


HUE_FACTOR = 360
PERCENT_FACTOR = 100


class HSB(ColorType):

    def __init__(self, hue: float, saturation: float, brightness: float):
        if not 0 <= hue <= HUE_FACTOR or not 0 <= saturation <= PERCENT_FACTOR or not 0 <= brightness <= PERCENT_FACTOR:
            raise ValueError()

        self._hue: float = hue
        self._saturation: float = saturation
        self._brightness: float = brightness

    @property
    def hue(self) -> float:
        """hue value"""
        return self._hue

    @property
    def h(self) -> float:
        """hue value"""
        return self._hue

    @property
    def saturation(self) -> float:
        """saturation value"""
        return self._saturation

    @property
    def s(self) -> float:
        """saturation value"""
        return self._saturation

    @property
    def brightness(self) -> float:
        """brightness value"""
        return self._brightness

    @property
    def b(self) -> float:
        """brightness value"""
        return self._brightness

    def replace(self, h: Optional[float] = None, s: Optional[float] = None, b: Optional[float] = None,
                hue: Optional[float] = None, saturation: Optional[float] = None,
                brightness: Optional[float] = None) -> Self:
        """Create a new object with (optionally) replaced values.

        :param h: New hue value
        :param hue: New hue value
        :param s: New saturation value
        :param saturation: New saturation value
        :param b: New brightness value
        :param brightness: New brightness value
        """

        if h is not None:
            if hue is not None:
                raise ValueError()
            hue = h

        if s is not None:
            if saturation is not None:
                raise ValueError()
            saturation = s

        if b is not None:
            if brightness is not None:
                raise ValueError()
            brightness = b

        return self.__class__(
            hue if hue is not None else self._hue,
            saturation if saturation is not None else self._saturation,
            brightness if brightness is not None else self._brightness,
        )

    def __str__(self):
        return f'{self.__class__.__name__}({self._hue:.2f}, {self._saturation:.2f}, {self._brightness:.2f})'

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._hue == other._hue and \
                self._saturation == other._saturation and \
                self._brightness == other._brightness
        else:
            return NotImplemented

    def __getitem__(self, item: int) -> float:
        if item == 0:
            return self._hue
        if item == 1:
            return self._saturation
        if item == 2:
            return self._brightness
        raise IndexError()

    # ------------------------------------------------------------------------------------------------------------------
    # Conversions
    # ------------------------------------------------------------------------------------------------------------------
    def to_rgb(self) -> RGB:
        """Create an RGB object from this object

        :return: New RGB object
        """
        return RGB.from_hsb(self)

    @classmethod
    def from_rgb(cls, obj: Union[RGB, Tuple[int, int, int]]) -> Self:
        """Create an HSB object from an RGB object or an RGB tuple

        :param obj: HSB object or RGB tuple
        :return: New HSB object
        """
        if not isinstance(obj, RGB):
            obj = RGB(*obj)
        return obj.to_hsb()



from typing import Any as __Any
from typing import Awaitable as __Awaitable
from typing import Callable as __Callable
from typing import Type as __Type

from .const import PYTHON_310 as __IS_GE_PYTHON_310

if __IS_GE_PYTHON_310:
    from typing import TypeAlias
else:
    from typing import Final as TypeAlias

HINT_ANY_CLASS: TypeAlias = __Type[object]
HINT_FUNC_ASYNC: TypeAlias = __Callable[..., __Awaitable[__Any]]

HINT_EVENT_CALLBACK: TypeAlias = __Callable[[__Any], __Any]



import logging
from queue import SimpleQueue, Empty
from threading import Thread
from time import sleep
from typing import Optional, Final

import HABApp
from .config import CONFIG

log = logging.getLogger('HABApp.logging')


class HABAppQueueHandler:
    FLUSH_DELAY: float = CONFIG.habapp.logging.flush_every

    def __init__(self, queue: SimpleQueue, handler_name: str, thread_name: str):
        self._handler: Optional[logging.Handler] = None
        self._handler_name: Final = handler_name
        self._queue: Final = queue
        self._name: Final = thread_name
        self._thread: Optional[Thread] = None

    def start(self) -> None:
        if self._thread is not None:
            raise RuntimeError('Thread can only be started once!')

        # resolve handler
        self._handler = logging._handlers[self._handler_name]

        self._thread = Thread(target=self._worker, name=f'HABApp_{self._name}')
        self._thread.start()

    def signal_stop(self):
        self._queue.put_nowait(None)

    def stop(self) -> None:
        if self._thread is None:
            return None
        thread = self._thread

        self.signal_stop()
        thread.join()

    def _worker(self):
        try:
            assert self._handler is not None
            while True:
                sleep(self.FLUSH_DELAY)
                if self.process_queue():
                    break

            log.debug(f'{self._name} thread stopped')
        except Exception as e:
            HABApp.core.wrapper.process_exception(self._worker, e)

        # clean up queue
        try:
            while True:
                self._queue.get_nowait()
        except Empty:
            pass

        self._thread = None

    def process_queue(self) -> bool:
        q = self._queue
        handler = self._handler

        first_rec = True

        check_interval = 200
        ctr = check_interval

        skip_rem = 0
        skip_total = 0
        skip_level = logging.INFO

        try:
            while True:
                if first_rec:
                    # first call is blocking
                    rec = q.get()           # type: Optional[logging.LogRecord]
                    first_rec = False
                else:
                    rec = q.get_nowait()    # type: Optional[logging.LogRecord]

                if rec is None:
                    return True

                if skip_rem > 0:
                    # skip everything including INFO, process rest
                    if rec.levelno <= skip_level:
                        skip_rem -= 1
                        if skip_rem <= 0:
                            log.warning(f'Event log buffer congested! Skipped {skip_total} messages.')
                        continue

                # handle record
                handler.handle(rec)

                ctr -= 1
                if ctr <= 0:
                    ctr = check_interval
                    if not skip_rem:
                        q_size = q.qsize()
                        if q_size > 1000:
                            skip_total = skip_rem = q_size - 750

        except Empty:
            return False



import typing

import voluptuous

_PARAMETERS: typing.Dict[str, dict] = {}
_VALIDATORS: typing.Dict[str, voluptuous.Schema] = {}


def remove_parameter_file(file):
    return _PARAMETERS.pop(file)


def set_parameter_file(file: str, value):
    # validate the parameters, this will raise an exception
    validator = _VALIDATORS.get(file)
    if validator is not None:
        value = validator(value)

    _PARAMETERS[file] = value


def get_parameter_file(file: str):
    return _PARAMETERS[file]


def set_file_validator(filename: str, validator: typing.Any, allow_extra_keys=True):
    """Add a validator for the parameter file. If the file is already loaded this will reload the file.

    :param filename: filename which shall be validated (without extension)
    :param validator: Description of file content - see the library
                      `voluptuous <https://github.com/alecthomas/voluptuous#show-me-an-example/>`_ for examples.
                      Use `None` to remove validator.
    :param allow_extra_keys: Allow additional keys in the file structure
    """

    # Remove validator
    if validator is None:
        _VALIDATORS.pop(filename, None)
        return

    # Set validator
    old_validator = _VALIDATORS.get(filename)
    _VALIDATORS[filename] = new_validator = voluptuous.Schema(
        validator, required=True, extra=(voluptuous.ALLOW_EXTRA if allow_extra_keys else voluptuous.PREVENT_EXTRA)
    )

    # todo: move this to file handling so we get the extension
    if old_validator != new_validator:
        reload_param_file(filename)


def add_parameter(file: str, *keys, default_value):
    save = False

    if file not in _PARAMETERS:
        save = True
        param: typing.Dict[str, typing.Any] = {}
        _PARAMETERS[file] = param
    else:
        param = _PARAMETERS[file]

    if keys:
        # Create structure
        for key in keys[:-1]:
            if key not in param:
                param[key] = {}
                save = True
            param = param[key]

        # Create value
        if keys[-1] not in param:
            param[keys[-1]] = default_value
            save = True

    if save:
        save_file(file)
    return None


def get_value(file: str, *keys) -> typing.Any:
    try:
        param = _PARAMETERS[file]
    except KeyError:
        raise FileNotFoundError(f'File {file}.yml not found in params folder!')

    # lookup parameter
    for key in keys:
        param = param[key]
    return param


# Import here to prevent cyclic imports
from .parameter_files import save_file, reload_param_file  # noqa: E402



from .base_item import BaseItem, HINT_TYPE_ITEM_OBJ, HINT_ITEM_OBJ
from .base_valueitem import BaseValueItem

# isort split

from .item import Item
from .item_color import ColorItem
from .item_aggregation import AggregationItem



from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import uses_get_item, uses_item_registry
from HABApp.core.items import BaseValueItem
from HABApp.mqtt.mqtt_interface import publish

get_item = uses_get_item()
item_registry = uses_item_registry()


class MqttBaseItem(BaseValueItem):
    pass


class MqttItem(MqttBaseItem):
    """A simple item that represents a topic and a value"""

    @classmethod
    def get_create_item(cls, name: str, initial_value=None) -> 'MqttItem':
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :param initial_value: state the item will have if it gets created
        :return: item
        """
        assert isinstance(name, str), type(name)

        try:
            item = get_item(name)
        except ItemNotFoundException:
            item = cls(name, initial_value)
            item_registry.add_item(item)

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def publish(self, payload, qos: int = None, retain: bool = None):
        """
        Publish the payload under the topic from the item.

        :param payload: MQTT Payload
        :param qos: QoS, can be ``0``, ``1`` or ``2``. If not specified value from configuration file will be used.
        :param retain: retain message. If not specified value from configuration file will be used.
        :return: 0 if successful
        """

        return publish(self.name, payload, qos=qos, retain=retain)



from typing import Set, Optional, Callable
from typing import TypeVar

from HABApp.core.errors import ContextBoundObjectIsAlreadyLinkedError, ContextBoundObjectIsAlreadyUnlinkedError


class ContextBoundObj:
    def __init__(self, parent_ctx: Optional['HINT_CONTEXT_OBJ'], **kwargs):
        super().__init__(**kwargs)
        self._parent_ctx = parent_ctx
        if parent_ctx is not None:
            parent_ctx.add_obj(self)

    def _ctx_link(self, parent_ctx: 'HINT_CONTEXT_OBJ'):
        assert isinstance(parent_ctx, Context)
        if self._parent_ctx is not None:
            raise ContextBoundObjectIsAlreadyLinkedError()

        self._parent_ctx = parent_ctx
        parent_ctx.add_obj(self)

    def _ctx_unlink(self):
        if self._parent_ctx is None:
            raise ContextBoundObjectIsAlreadyUnlinkedError()

        self._parent_ctx.remove_obj(self)
        self._parent_ctx = None


HINT_CONTEXT_BOUND_OBJ = TypeVar('HINT_CONTEXT_BOUND_OBJ', bound=ContextBoundObj)


class Context:
    def __init__(self):
        self.objs: Set[HINT_CONTEXT_BOUND_OBJ] = set()

    def add_obj(self, obj: HINT_CONTEXT_BOUND_OBJ):
        assert isinstance(obj, ContextBoundObj)
        self.objs.add(obj)

    def remove_obj(self, obj: HINT_CONTEXT_BOUND_OBJ):
        assert isinstance(obj, ContextBoundObj)
        self.objs.remove(obj)

    def link(self, obj: HINT_CONTEXT_BOUND_OBJ) -> HINT_CONTEXT_BOUND_OBJ:
        assert isinstance(obj, ContextBoundObj)
        obj._ctx_link(self)
        return obj

    def get_callback_name(self, callback: Callable) -> Optional[str]:
        raise NotImplementedError()


HINT_CONTEXT_OBJ = TypeVar('HINT_CONTEXT_OBJ', bound=Context)


class ContextProvidingObj:
    def __init__(self, context: Optional[HINT_CONTEXT_OBJ] = None, **kwargs):
        super().__init__(**kwargs)
        self._habapp_ctx: HINT_CONTEXT_OBJ = context



import asyncio
import logging
import time
from typing import Optional

import HABApp
from HABApp.core.internals import uses_event_bus
from HABApp.core.wrapper import log_exception
from HABApp.openhab.errors import OpenhabDisconnectedError
from ._plugin import PluginBase

log = logging.getLogger('HABApp.openhab.ping')


event_bus = uses_event_bus()


class PingOpenhab(PluginBase):
    def __init__(self):
        self.ping_value: Optional[float] = None
        self.ping_sent: Optional[float] = None
        self.ping_new: Optional[float] = None

        self.listener: Optional[HABApp.core.internals.EventBusListener] = None

        self.fut_ping: Optional[asyncio.Future] = None

    def setup(self):
        HABApp.config.CONFIG.openhab.ping.subscribe_for_changes(self.cfg_changed)
        self.cfg_changed()

    def on_connect(self):
        if not self.IS_CONNECTED:
            return None

        if not HABApp.config.CONFIG.openhab.ping.enabled:
            return None

        # initialize
        self.ping_value = None
        self.ping_sent = None
        self.ping_new = None

        self.fut_ping = asyncio.create_task(self.async_ping())

    def on_disconnect(self):
        if self.fut_ping is not None:
            self.fut_ping.cancel()
            self.fut_ping = None

        log.debug('Ping stopped')

    def cfg_changed(self):
        self.on_disconnect()

        if self.listener is not None:
            self.listener.cancel()
            self.listener = None

        if not HABApp.config.CONFIG.openhab.ping.enabled:
            return None

        self.listener = HABApp.core.internals.EventBusListener(
            HABApp.config.CONFIG.openhab.ping.item,
            HABApp.core.internals.wrap_func(self.ping_received),
            HABApp.core.events.EventFilter(HABApp.openhab.events.ItemStateEvent)
        )
        event_bus.add_listener(self.listener)

        self.on_connect()

    async def ping_received(self, event: HABApp.openhab.events.ItemStateEvent):
        value = event.value
        if value != self.ping_value:
            return None

        # We only save take the first ping we get
        if self.ping_new is not None:
            return None

        self.ping_new = round((time.time() - self.ping_sent) * 1000, 1)


    @log_exception
    async def async_ping(self):
        await asyncio.sleep(3)

        log.debug('Ping started')
        try:
            while True:

                self.ping_value = self.ping_new
                self.ping_new = None
                self.ping_sent = time.time()

                await HABApp.openhab.interface_async.async_post_update(
                    HABApp.config.CONFIG.openhab.ping.item,
                    f'{self.ping_value:.1f}' if self.ping_value is not None else None
                )

                await asyncio.sleep(HABApp.config.CONFIG.openhab.ping.interval)

        except OpenhabDisconnectedError:
            pass


PLUGIN_PING = PingOpenhab.create_plugin()



import datetime
import typing
import warnings
from typing import Any, Optional, Dict, List
from urllib.parse import quote as quote_url

from pydantic import parse_obj_as

from HABApp.core.const.json import load_json
from HABApp.core.items import BaseValueItem
from HABApp.openhab.definitions.rest import ItemChannelLinkDefinition, LinkNotFoundError, OpenhabThingDefinition
from HABApp.openhab.definitions.rest.habapp_data import get_api_vals, load_habapp_meta
from HABApp.openhab.errors import ThingNotEditableError, \
    ThingNotFoundError, ItemNotEditableError, ItemNotFoundError, MetadataNotEditableError
from .http_connection import delete, get, put, post, async_get_root, async_get_uuid, async_send_command, \
    async_post_update
from HABApp.core.types import HSB, RGB

if typing.TYPE_CHECKING:
    post = post
    async_get_root = async_get_root
    async_get_uuid = async_get_uuid
    async_send_command = async_send_command
    async_post_update = async_post_update


def convert_to_oh_type(_in: Any) -> str:
    if isinstance(_in, BaseValueItem):
        raise ValueError()

    if isinstance(_in, datetime.datetime):
        # Add timezone (if not yet defined) to string, then remote anything below ms.
        # 2018-11-19T09:47:38.284000+0100 -> 2018-11-19T09:47:38.284+0100
        out = _in.astimezone(None).strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        return out

    if isinstance(_in, (set, list, tuple, frozenset)):
        return ','.join(map(str, _in))

    if _in is None:
        return 'NULL'

    if isinstance(_in, RGB):
        _in = _in.to_hsb()

    if isinstance(_in, HSB):
        return f'{_in._hue:.2f},{_in._saturation:.2f},{_in._brightness:.2f}'

    return str(_in)


async def async_item_exists(item) -> bool:
    ret = await get(f'/rest/items/{item:s}', log_404=False)
    return ret.status == 200


async def async_get_items(include_habapp_meta=False, metadata: Optional[str] = None,
                          all_metadata=False) -> Optional[List[Dict[str, Any]]]:
    params = None
    if include_habapp_meta:
        params = {'metadata': 'HABApp'}
    if metadata is not None:
        if params is not None:
            raise ValueError('Use include_habapp_meta or metadata')
        params = {'metadata': metadata}
    if all_metadata:
        params = {'metadata': '.+'}

    resp = await get('/rest/items', params=params)
    return await resp.json(loads=load_json, encoding='utf-8')


async def async_get_item(item: str, metadata: Optional[str] = None, all_metadata=False) -> dict:
    params = None if metadata is None else {'metadata': metadata}
    if all_metadata:
        params = {'metadata': '.+'}

    ret = await get(f'/rest/items/{item:s}', params=params, log_404=False)
    if ret.status == 404:
        raise ItemNotFoundError.from_name(item)
    if ret.status >= 300:
        return {}
    else:
        data = await ret.json(loads=load_json, encoding='utf-8')
        return data


async def async_get_things() -> List[OpenhabThingDefinition]:
    resp = await get('/rest/things')
    data = await resp.json(loads=load_json, encoding='utf-8')

    return parse_obj_as(List[OpenhabThingDefinition], data)


async def async_get_thing(uid: str) -> OpenhabThingDefinition:
    ret = await get(f'/rest/things/{uid:s}')
    if ret.status >= 300:
        raise ThingNotFoundError.from_uid(uid)

    return OpenhabThingDefinition.parse_obj(await ret.json(loads=load_json, encoding='utf-8'))


async def async_get_persistence_data(item_name: str, persistence: typing.Optional[str],
                                     start_time: typing.Optional[datetime.datetime],
                                     end_time: typing.Optional[datetime.datetime]) -> dict:

    params = {}
    if persistence:
        params['serviceId'] = persistence
    if start_time is not None:
        params['starttime'] = convert_to_oh_type(start_time)
    if end_time is not None:
        params['endtime'] = convert_to_oh_type(end_time)
    if not params:
        params = None

    ret = await get(f'/rest/persistence/items/{item_name:s}', params=params)
    if ret.status >= 300:
        return {}
    else:
        return await ret.json(loads=load_json, encoding='utf-8')


async def async_set_persistence_data(item_name: str, persistence: typing.Optional[str],
                                     time: datetime.datetime, state: typing.Any):

    # This does not work as of OH 3.2
    warnings.warn(f'{async_set_persistence_data.__name__} calls a part of the openHAB API which is buggy!',
                  category=ResourceWarning)

    params = {
        'itemname': item_name,
        'time': convert_to_oh_type(time),
        'state': convert_to_oh_type(state),
    }
    if persistence is not None:
        params['serviceId'] = persistence

    ret = await put(f'/rest/persistence/items/{item_name:s}', params=params)
    if ret.status >= 300:
        return {}
    else:
        return await ret.json(loads=load_json, encoding='utf-8')


async def async_create_item(item_type, name, label="", category="", tags=[], groups=[],
                            group_type=None, group_function=None, group_function_params=[]) -> bool:

    payload = {'type': item_type, 'name': name}
    if label:
        payload['label'] = label
    if category:
        payload['category'] = category
    if tags:
        payload['tags'] = tags
    if groups:
        payload['groupNames'] = groups  # CamelCase!

    # we create a group
    if group_type:
        payload['groupType'] = group_type   # CamelCase!
    if group_function:
        payload['function'] = {}
        payload['function']['name'] = group_function
        if group_function_params:
            payload['function']['params'] = group_function_params

    ret = await put(f'/rest/items/{name:s}', json=payload)
    if ret is None:
        return False

    if ret.status == 404:
        raise ItemNotFoundError.from_name(name)
    elif ret.status == 405:
        raise ItemNotEditableError.from_name(name)
    return ret.status < 300


async def async_remove_item(item):
    await delete(f'/rest/items/{item:s}')


async def async_remove_metadata(item: str, namespace: str):
    ret = await delete(f'/rest/items/{item:s}/metadata/{namespace:s}')
    if ret is None:
        return False

    if ret.status == 404:
        raise ItemNotFoundError.from_name(item)
    elif ret.status == 405:
        raise MetadataNotEditableError.create_text(item, namespace)
    return ret.status < 300


async def async_set_metadata(item: str, namespace: str, value: str, config: dict):
    payload = {
        'value': value,
        'config': config
    }
    ret = await put(f'/rest/items/{item:s}/metadata/{namespace:s}', json=payload)
    if ret is None:
        return False

    if ret.status == 404:
        raise ItemNotFoundError.from_name(item)
    elif ret.status == 405:
        raise MetadataNotEditableError.create_text(item, namespace)
    return ret.status < 300


async def async_set_thing_cfg(uid: str, cfg: typing.Dict[str, typing.Any]):
    ret = await put(f'/rest/things/{uid:s}/config', json=cfg)
    if ret is None:
        return None

    if ret.status == 404:
        raise ThingNotFoundError.from_uid(uid)
    elif ret.status == 409:
        raise ThingNotEditableError.from_uid(uid)
    elif ret.status >= 300:
        raise ValueError('Something went wrong')

    return ret.status


async def async_set_thing_enabled(uid: str, enabled: bool):
    ret = await put(f'/rest/things/{uid:s}/enable', data='true' if enabled else 'false')
    if ret is None:
        return None

    if ret.status == 404:
        raise ThingNotFoundError.from_uid(uid)
    elif ret.status == 409:
        raise ThingNotEditableError.from_uid(uid)
    elif ret.status >= 300:
        raise ValueError('Something went wrong')

    return ret.status


# ---------------------------------------------------------------------------------------------------------------------
# Link handling is experimental
# ---------------------------------------------------------------------------------------------------------------------

def __get_link_url(channel_uid: str, item_name: str) -> str:
    # rest/links/ endpoint needs the channel to be url encoded
    # (AAAA:BBBB:CCCC:0#NAME -> AAAA%3ABBBB%3ACCCC%3A0%23NAME)
    # otherwise the REST-api returns HTTP-Status 500 InternalServerError
    return '/rest/links/' + quote_url(f"{item_name}/{channel_uid}")


async def async_remove_channel_link(channel_uid: str, item_name: str) -> bool:
    ret = await delete(__get_link_url(channel_uid, item_name))
    if ret is None:
        return False
    return ret.status == 200


async def async_get_channel_links() -> List[Dict[str, str]]:
    ret = await get('/rest/links')
    if ret.status >= 300:
        return None
    else:
        return await ret.json(loads=load_json, encoding='utf-8')


async def async_get_channel_link_mode_auto() -> bool:
    ret = await get('/rest/links/auto')
    if ret.status >= 300:
        return False
    else:
        return await ret.json(loads=load_json, encoding='utf-8')


async def async_get_channel_link(channel_uid: str, item_name: str) -> ItemChannelLinkDefinition:
    ret = await get(__get_link_url(channel_uid, item_name), log_404=False)
    if ret.status == 404:
        raise LinkNotFoundError(f'Link {item_name} -> {channel_uid} not found!')
    if ret.status >= 300:
        return None
    else:
        return ItemChannelLinkDefinition(**await ret.json(loads=load_json, encoding='utf-8'))


async def async_channel_link_exists(channel_uid: str, item_name: str) -> bool:
    ret = await get(__get_link_url(channel_uid, item_name), log_404=False)
    return ret.status == 200


async def async_create_channel_link(
        channel_uid: str, item_name: str, configuration: Optional[Dict[str, Any]] = None) -> bool:

    # if the passed item doesn't exist OpenHAB creates a new empty item item
    # this is undesired and why we raise an Exception
    if not await async_item_exists(item_name):
        raise ItemNotFoundError.from_name(item_name)

    ret = await put(
        __get_link_url(channel_uid, item_name),
        json={'configuration': configuration} if configuration is not None else {}
    )
    if ret is None:
        return False
    return ret.status == 200


# ---------------------------------------------------------------------------------------------------------------------
# Funcs for handling HABApp Metadata
# ---------------------------------------------------------------------------------------------------------------------
async def async_remove_habapp_metadata(item: str):
    return await async_remove_metadata(item, 'HABApp')


async def async_set_habapp_metadata(item: str, obj):
    val, cfg = get_api_vals(obj)
    return await async_set_metadata(item, 'HABApp', val, cfg)


async def async_get_item_with_habapp_meta(item: str) -> dict:
    data = await async_get_item(item, metadata='HABApp')
    return load_habapp_meta(data)



from enum import Enum as _Enum
import typing


def _get_str_enum_values(obj: typing.Type[_Enum]) -> typing.Set[str]:
    return {_member.value for _member in obj}


class ItemType(str, _Enum):
    STRING = 'String'
    NUMBER = 'Number'
    SWITCH = 'Switch'
    CONTACT = 'Contact'
    DIMMER = 'Dimmer'
    ROLLERSHUTTER = 'Rollershutter'
    COLOR = 'Color'
    DATETIME = 'DateTime'
    LOCATION = 'Location'
    PLAYER = 'Player'
    GROUP = 'Group'
    IMAGE = 'Image'
    CALL = 'Call'


ITEM_TYPES: typing.Final = _get_str_enum_values(ItemType)


class ItemDimensions(str, _Enum):
    ACCELERATION = 'Acceleration'
    ANGLE = 'Angle'
    AREAL_DENSITY = 'ArealDensity'
    CATALYTIC_ACTIVITY = 'CatalyticActivity'
    DATA_AMOUNT = 'DataAmount'
    DATA_TRANSFER_RATE = 'DataTransferRate'
    DENSITY = 'Density'
    DIMENSIONLESS = 'Dimensionless'
    ELECTRIC_CAPACITANCE = 'ElectricCapacitance'
    ELECTRIC_CHARGE = 'ElectricCharge'
    ELECTRIC_CONDUCTANCE = 'ElectricConductance'
    ELECTRIC_CONDUCTIVITY = 'ElectricConductivity'
    ELECTRIC_CURRENT = 'ElectricCurrent'
    ELECTRIC_INDUCTANCE = 'ElectricInductance'
    ELECTRIC_POTENTIAL = 'ElectricPotential'
    ELECTRIC_RESISTANCE = 'ElectricResistance'
    ENERGY = 'Energy'
    FORCE = 'Force'
    FREQUENCY = 'Frequency'
    ILLUMINANCE = 'Illuminance'
    INTENSITY = 'Intensity'
    LENGTH = 'Length'
    LUMINOUS_FLUX = 'LuminousFlux'
    LUMINOUS_INTENSITY = 'LuminousIntensity'
    MAGNETIC_FLUX = 'MagneticFlux'
    MAGNETIC_FLUX_DENSITY = 'MagneticFluxDensity'
    MASS = 'Mass'
    POWER = 'Power'
    PRESSURE = 'Pressure'
    RADIATION_DOSE_ABSORBED = 'RadiationDoseAbsorbed'
    RADIATION_DOSE_EFFECTIVE = 'RadiationDoseEffective'
    RADIOACTIVITY = 'Radioactivity'
    SOLID_ANGLE = 'SolidAngle'
    SPEED = 'Speed'
    TEMPERATURE = 'Temperature'
    TIME = 'Time'
    VOLUME = 'Volume'
    VOLUMETRIC_FLOW_RATE = 'VolumetricFlowRate'


ITEM_DIMENSIONS: typing.Final = _get_str_enum_values(ItemDimensions)


class GroupItemFunctions(str, _Enum):
    AND = 'AND'
    AVG = 'AVG'
    MAX = 'MAX'
    MIN = 'MIN'
    NAND = 'NAND'
    NOR = 'NOR'
    OR = 'OR'
    SUM = 'SUM'


GROUP_ITEM_FUNCTIONS: typing.Final = _get_str_enum_values(GroupItemFunctions)



from .base_event import OpenhabEvent
from .item_events import ItemStateEvent, ItemStateChangedEvent, ItemCommandEvent, ItemAddedEvent,\
    ItemUpdatedEvent, ItemRemovedEvent, ItemStatePredictedEvent, GroupItemStateChangedEvent
from .channel_events import ChannelTriggeredEvent, ChannelDescriptionChangedEvent
from .thing_events import ThingStatusInfoChangedEvent, ThingStatusInfoEvent, \
    ThingFirmwareStatusInfoEvent, ThingAddedEvent, ThingRemovedEvent, ThingUpdatedEvent, ThingConfigStatusInfoEvent
from .event_filters import ItemStateEventFilter, ItemStateChangedEventFilter, ItemCommandEventFilter



from HABApp.openhab.connection_handler import http_connection
from ._plugin import on_connect, on_disconnect, setup_plugins

log = http_connection.log


def setup():
    from HABApp.runtime import shutdown

    # initialize callbacks
    http_connection.ON_CONNECTED = on_connect
    http_connection.ON_DISCONNECTED = on_disconnect

    # shutdown handler for connection
    shutdown.register_func(http_connection.shutdown_connection, msg='Stopping openHAB connection')

    # shutdown handler for plugins
    shutdown.register_func(on_disconnect, msg='Stopping openHAB plugins')

    # initialize all plugins
    setup_plugins()
    return None


async def start():
    await http_connection.setup_connection()



from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import uses_item_registry, uses_get_item
from HABApp.core.items import BaseValueItem

get_item = uses_get_item()
item_registry = uses_item_registry()


class Item(BaseValueItem):
    """Simple item, used to store values in HABApp"""

    @classmethod
    def get_create_item(cls, name: str, initial_value=None) -> 'Item':
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name

        :param name: item name
        :param initial_value: state the item will have if it gets created
        :return: The item
        """
        assert isinstance(name, str), type(name)

        try:
            item = get_item(name)
        except ItemNotFoundException:
            item = cls(name, initial_value)
            item_registry.add_item(item)

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item



import typing
from math import ceil, floor

from .parameters import add_parameter as _add_parameter
from .parameters import get_value as _get_value


class BaseParameter:
    def __init__(self, filename: str, *keys, default_value: typing.Any = 'ToDo'):
        """Class to dynamically access parameters which are loaded from file.

        :param filename: filename (without extension)
        :param keys: structure in the file
        :param default_value: default value for the parameter. Is used to create the file and the structure if
                              it does not exist yet. Use ``None`` to skip creation of the file structure.
        """

        assert isinstance(filename, str), type(filename)
        self._filename: str = filename
        self._keys = keys

        # as a convenience try to create the file and the file structure
        if default_value is not None:
            _add_parameter(self._filename, *self._keys, default_value=default_value)


class Parameter(BaseParameter):

    @property
    def value(self) -> typing.Any:
        """Return the current value. This will do the lookup so make sure to not cache this value, otherwise
        the parameter might not work as expected.
        """
        return _get_value(self._filename, *self._keys)

    def __repr__(self):
        return f'<Parameter file: {self._filename}, keys: {self._keys}, value: {self.value}'

    def __bool__(self):
        return bool(self.value)

    def __eq__(self, other):
        return self.value == other

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __ge__(self, other):
        return self.value >= other

    def __gt__(self, other):
        return self.value > other

    # https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types
    # These methods are called to implement the binary arithmetic operations
    def __add__(self, other):
        return self.value.__add__(other)

    def __sub__(self, other):
        return self.value.__sub__(other)

    def __mul__(self, other):
        return self.value.__mul__(other)

    def __matmul__(self, other):
        return self.value.__matmul__(other)

    def __truediv__(self, other):
        return self.value.__truediv__(other)

    def __floordiv__(self, other):
        return self.value.__floordiv__(other)

    def __mod__(self, other):
        return self.value.__mod__(other)

    def __divmod__(self, other):
        return self.value.__divmod__(other)

    def __pow__(self, other):
        return self.value.__pow__(other)

    def __lshift__(self, other):
        return self.value.__lshift__(other)

    def __rshift__(self, other):
        return self.value.__rshift__(other)

    def __and__(self, other):
        return self.value.__and__(other)

    def __xor__(self, other):
        return self.value.__xor__(other)

    def __or__(self, other):
        return self.value.__or__(other)

    # Unary arithmetic operations (-, +, abs() and ~).
    def __neg__(self):
        return self.value.__neg__()

    def __pos__(self):
        return self.value.__pos__()

    def __abs__(self):
        return self.value.__abs__()

    def __invert__(self):
        return self.value.__invert__()

    # built-in functions complex(), int() and float().
    def __complex__(self):
        return self.value.__complex__()

    def __int__(self):
        return self.value.__int__()

    def __float__(self):
        return self.value.__float__()

    # built-in function round() and math functions trunc(), floor() and ceil().
    def __round__(self, ndigits=None):
        return self.value.__round__(ndigits)

    def __trunc__(self):
        return self.value.__trunc__()

    def __floor__(self):
        return floor(self.value)

    def __ceil__(self):
        return ceil(self.value)

    # we don't support modification in place! We have to override this because otherwise
    # python falls back to the methods above
    def __iadd__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __isub__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __imul__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __imatmul__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __itruediv__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __ifloordiv__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __imod__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __ipow__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __ilshift__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __irshift__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __iand__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __ixor__(self, other):
        raise PermissionError('Parameter can not be changed!')

    def __ior__(self, other):
        raise PermissionError('Parameter can not be changed!')


class DictParameter(BaseParameter):
    """Implements a dict interface"""

    @property
    def value(self) -> dict:
        """Return the current value. This will do the lookup so make sure to not cache this value, otherwise
        the parameter might not work as expected.
        """
        value = _get_value(self._filename, *self._keys)
        if not isinstance(value, dict):
            raise ValueError(f'Value "{value}" for {self.__class__.__name__} is not a dict! ({type(value)})')
        return value

    def __repr__(self):
        return f'<DictParameter file: {self._filename}, keys: {self._keys}, value: {self.value}'

    def __bool__(self):
        return bool(self.value)

    def __eq__(self, other):
        return self.value == other

    def __getitem__(self, item):
        return self.value[item]

    def __contains__(self, key):
        return key in self.value

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def keys(self):
        return self.value.keys()

    def values(self):
        return self.value.values()

    def items(self):
        return self.value.items()

    def get(self, item, default=None):
        return self.value.get(item, default)

    def __setitem__(self, key, value):
        raise PermissionError('Parameter can not be changed!')

    def __delitem__(self, key):
        raise PermissionError('Parameter can not be changed!')



import logging
from pathlib import Path
from threading import Lock
from typing import Optional, Dict

from watchdog.observers import Observer
from watchdog.observers.api import ObservedWatch

from .base_watcher import FileSystemEventHandler

log = logging.getLogger('HABApp.files.watcher')

LOCK = Lock()

OBSERVER: Optional[Observer] = None
WATCHES: Dict[str, ObservedWatch] = {}


def start():
    global OBSERVER

    # start only once!
    assert OBSERVER is None

    OBSERVER = Observer()
    OBSERVER.start()

    # register for proper shutdown
    from HABApp.runtime import shutdown
    shutdown.register_func(OBSERVER.stop, msg='Stopping folder observer')
    shutdown.register_func(OBSERVER.join, last=True, msg='Joining folder observer threads')
    return None


def add_folder_watch(handler: FileSystemEventHandler):
    assert OBSERVER is not None
    assert isinstance(handler, FileSystemEventHandler), type(handler)
    assert isinstance(handler.folder, Path) and handler.folder.is_dir()

    log.debug(
        f'Adding {"recursive " if handler.watch_subfolders else ""}watcher for {handler.folder} with {handler.filter}'
    )

    with LOCK:
        _folder = str(handler.folder)
        assert _folder not in WATCHES

        WATCHES[_folder] = OBSERVER.schedule(handler, _folder, recursive=handler.watch_subfolders)


def remove_folder_watch(folder: Path):
    assert OBSERVER is not None
    assert isinstance(folder, Path)

    with LOCK:
        OBSERVER.unschedule(WATCHES.pop(str(folder)))



import logging
from typing import Optional, TypeVar, Callable

from HABApp.core.const.topics import TOPIC_ERRORS as TOPIC_ERRORS
from HABApp.core.events.habapp_events import HABAppException
from HABApp.core.internals import HINT_CONTEXT_OBJ, ContextProvidingObj, uses_event_bus
from HABApp.core.lib import format_exception

default_logger = logging.getLogger('HABApp.Worker')

event_bus = uses_event_bus()


class WrappedFunctionBase(ContextProvidingObj):

    def __init__(self, func: Callable, name: Optional[str] = None, logger: Optional[logging.Logger] = None,
                 context: Optional[HINT_CONTEXT_OBJ] = None):

        # Allow setting of the rule context
        super().__init__(context)

        # name of the function
        if name is None:
            if self._habapp_ctx is not None:
                name = self._habapp_ctx.get_callback_name(func)
            if name is None:
                name = func.__name__
        self.name: str = name

        # Allow custom logger
        self.log = default_logger
        if logger:
            self.log = logger

    def run(self, *args, **kwargs):
        raise NotImplementedError()

    def process_exception(self, e: Exception, *args, **kwargs):

        lines = format_exception(e)

        # Log Exception
        self.log.error(f'Error in {self.name}: {e}')
        for line in lines:
            self.log.error(line)

        # Create HABApp event, but only if we are not currently processing an exception while processing an error.
        # Otherwise we might create an endless loop!
        if not args or not isinstance(args[0], HABAppException):
            event_bus.post_event(
                TOPIC_ERRORS, HABAppException(func_name=self.name, exception=e, traceback='\n'.join(lines))
            )


TYPE_WRAPPED_FUNC_OBJ = TypeVar('TYPE_WRAPPED_FUNC_OBJ', bound=WrappedFunctionBase)



from typing import Any, Dict, Optional, Tuple

from pydantic import BaseModel, Field, validator


class OpenhabThingChannelDefinition(BaseModel):
    uid: str
    id: str
    channel_type: str = Field(..., alias='channelTypeUID')
    kind: str

    label: str = ''
    description: str = ''
    item_type: Optional[str] = Field(None, alias='itemType')

    linked_items: Tuple[str, ...] = Field(default_factory=tuple, alias='linkedItems')
    default_tags: Tuple[str, ...] = Field(default_factory=tuple, alias='defaultTags')

    properties: Dict[str, Any] = Field(default_factory=dict)
    configuration: Dict[str, Any] = Field(default_factory=dict)


class OpenhabThingStatus(BaseModel):
    status: str
    detail: str = Field(..., alias='statusDetail')
    description: Optional[str] = None

    @validator('detail')
    def _parse_detail(cls, v):
        return '' if v == 'NONE' else v


class OpenhabThingDefinition(BaseModel):
    # These are mandatory fields
    editable: bool
    status: OpenhabThingStatus = Field(..., alias='statusInfo')
    thing_type: str = Field(..., alias='thingTypeUID')
    uid: str = Field(..., alias='UID')

    # These fields are optional, but we want to have a value set
    # because it simplifies the thing handling a lot
    bridge_uid: Optional[str] = Field(None, alias='bridgeUID')
    label: str = ''
    location: str = ''

    # Containers should always have a default, so it's easy to iterate over them
    channels: Tuple[OpenhabThingChannelDefinition, ...] = Field(default_factory=tuple)
    configuration: Dict[str, Any] = Field(default_factory=dict)
    firmwareStatus: Dict[str, str] = Field(default_factory=dict)
    properties: Dict[str, Any] = Field(default_factory=dict)



class __FileEventBase:
    def __init__(self, name: str):
        self.name: str = name

    def __repr__(self):
        return f'<{self.__class__.__name__} filename: {self.name}>'


class RequestFileLoadEvent(__FileEventBase):
    """Request (re-) loading of the specified file

    :ivar str filename: relative filename
    """


class RequestFileUnloadEvent(__FileEventBase):
    """Request unloading of the specified file

    :ivar str filename: relative filename
    """


class HABAppException:
    """Contains information about an Exception that has occurred in HABApp

    :ivar str func_name: name of the function where the error occurred
    :ivar str traceback: traceback
    :ivar Exception exception: Exception
    """
    def __init__(self, func_name: str, exception: Exception, traceback: str):
        self.func_name: str = func_name
        self.exception: Exception = exception
        self.traceback: str = traceback

    def __repr__(self):
        return f'<{self.__class__.__name__} func_name: {self.func_name}, exception: {self.exception}>'

    def to_str(self) -> str:
        """Create a readable str with all information"""
        return f'Exception in {self.func_name}: {self.exception}\n{self.traceback}'



import asyncio
import logging
import traceback
import typing
from asyncio import Queue, sleep, QueueEmpty
from typing import Any, Optional, Final

import aiohttp
from aiohttp.client import ClientResponse, _RequestContextManager
from aiohttp.hdrs import METH_GET, METH_POST, METH_PUT, METH_DELETE
from aiohttp_sse_client import client as sse_client

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core.asyncio import async_context
from HABApp.core.const.json import dump_json, load_json
from HABApp.core.logger import log_info, log_warning
from HABApp.core.wrapper import process_exception, ignore_exception
from HABApp.openhab.errors import OpenhabDisconnectedError, ExpectedSuccessFromOpenhab
from .http_connection_waiter import WaitBetweenConnects
from ...core.const.topics import TOPIC_EVENTS
from ...core.lib import SingleTask

log = logging.getLogger('HABApp.openhab.connection')
log_events = logging.getLogger(f'{TOPIC_EVENTS}.openhab')


IS_ONLINE: bool = False
IS_READ_ONLY: bool = False


# HTTP options
HTTP_ALLOW_REDIRECTS: bool = True
HTTP_VERIFY_SSL: Optional[bool] = None
HTTP_SESSION: aiohttp.ClientSession = None

CONNECT_WAIT: WaitBetweenConnects = WaitBetweenConnects()

ON_CONNECTED: typing.Callable = None
ON_DISCONNECTED: typing.Callable = None


async def get(url: str, log_404=True, **kwargs: Any) -> ClientResponse:

    mgr = _RequestContextManager(
        HTTP_SESSION._request(METH_GET, url, allow_redirects=HTTP_ALLOW_REDIRECTS, ssl=HTTP_VERIFY_SSL, **kwargs)
    )
    return await check_response(mgr, log_404=log_404)


async def post(url: str, log_404=True, json=None, data=None, **kwargs: Any) -> Optional[ClientResponse]:

    if IS_READ_ONLY or not IS_ONLINE:
        return None

    mgr = _RequestContextManager(
        HTTP_SESSION._request(
            METH_POST, url, allow_redirects=HTTP_ALLOW_REDIRECTS, ssl=HTTP_VERIFY_SSL,
            data=data, json=json, **kwargs
        )
    )

    if data is None:
        data = json
    return await check_response(mgr, log_404=log_404, sent_data=data)


async def put(url: str, log_404=True, json=None, data=None, **kwargs: Any) -> Optional[ClientResponse]:

    if IS_READ_ONLY or not IS_ONLINE:
        return None

    mgr = _RequestContextManager(
        HTTP_SESSION._request(
            METH_PUT, url, allow_redirects=HTTP_ALLOW_REDIRECTS, ssl=HTTP_VERIFY_SSL,
            data=data, json=json, **kwargs
        )
    )

    if data is None:
        data = json
    return await check_response(mgr, log_404=log_404, sent_data=data)


async def delete(url: str, log_404=True, json=None, data=None, **kwargs: Any) -> Optional[ClientResponse]:

    if IS_READ_ONLY or not IS_ONLINE:
        return None

    mgr = _RequestContextManager(
        HTTP_SESSION._request(METH_DELETE, url, allow_redirects=HTTP_ALLOW_REDIRECTS, ssl=HTTP_VERIFY_SSL,
                              data=data, json=json, **kwargs)
    )

    if data is None:
        data = json
    return await check_response(mgr, log_404=log_404, sent_data=data)


def set_offline(log_msg=''):
    global IS_ONLINE

    if not IS_ONLINE:
        return None
    IS_ONLINE = False

    log.warning(f'Disconnected! {log_msg}')

    # cancel SSE listener
    TASK_SSE_LISTENER.cancel()
    TASK_TRY_CONNECT.cancel()

    ON_DISCONNECTED()

    TASK_TRY_CONNECT.start()


def is_disconnect_exception(e) -> bool:
    if not isinstance(e, (
            # aiohttp Exceptions
            aiohttp.ClientPayloadError, aiohttp.ClientConnectorError, aiohttp.ClientOSError,

            # aiohttp_sse_client Exceptions
            ConnectionRefusedError, ConnectionError, ConnectionAbortedError)):
        return False

    set_offline(str(e))
    return True


async def check_response(future: aiohttp.client._RequestContextManager, sent_data=None,
                         log_404=True, disconnect_on_error=False) -> ClientResponse:
    try:
        resp = await future
    except Exception as e:
        is_disconnect = is_disconnect_exception(e)
        log.log(logging.WARNING if is_disconnect else logging.ERROR, f'"{e}" ({type(e)})')
        if is_disconnect:
            raise OpenhabDisconnectedError()
        raise

    status = resp.status

    # Sometimes openHAB issues 404 instead of 500 during startup
    if disconnect_on_error and status >= 400:
        set_offline(f'Expected success but got status {status} for '
                    f'{str(resp.request_info.url).replace(HABApp.CONFIG.openhab.connection.url, "")}')
        raise ExpectedSuccessFromOpenhab()

    # Something went wrong - log error message
    log_msg = False
    if status >= 300:
        log_msg = True

        # possibility skip logging of 404
        if status == 404 and not log_404:
            log_msg = False

    if log_msg:
        # Log Error Message
        sent = '' if sent_data is None else f' {sent_data}'
        log.warning(f'Status {status} for {resp.request_info.method} {resp.request_info.url}{sent}')
        for line in str(resp).splitlines():
            log.debug(line)

    return resp


async def shutdown_connection():
    global HTTP_SESSION

    TASK_TRY_CONNECT.cancel()
    TASK_SSE_LISTENER.cancel()

    TASK_QUEUE_WORKER.cancel()
    TASK_QUEUE_WATCHER.cancel()

    await asyncio.sleep(0)

    # If we are already connected properly disconnect
    if HTTP_SESSION is not None:
        await HTTP_SESSION.close()
        HTTP_SESSION = None


async def setup_connection():
    global HTTP_SESSION, HTTP_VERIFY_SSL

    await shutdown_connection()

    config = HABApp.CONFIG.openhab
    url: str = config.connection.url

    # do not run without an url
    if not url:
        log_info(log, 'openHAB connection disabled!')
        return None

    if not config.connection.verify_ssl:
        HTTP_VERIFY_SSL = False
        log.info('Verify ssl set to False!')
    else:
        HTTP_VERIFY_SSL = None

    HTTP_SESSION = aiohttp.ClientSession(
        base_url=url,
        timeout=aiohttp.ClientTimeout(total=None),
        json_serialize=dump_json,
        auth=aiohttp.BasicAuth(config.connection.user, config.connection.password),
        read_bufsize=config.connection.buffer,
    )

    TASK_TRY_CONNECT.start()


async def start_sse_event_listener():

    async_context.set('SSE')

    try:
        # cache so we don't have to look up every event
        _load_json = load_json
        _see_handler = on_sse_event

        async with sse_client.EventSource(url=f'/rest/events?topics={HABApp.CONFIG.openhab.connection.topic_filter}',
                                          session=HTTP_SESSION, ssl=HTTP_VERIFY_SSL) as event_source:
            async for event in event_source:

                e_str = event.data

                try:
                    e_json = _load_json(e_str)
                except (ValueError, TypeError):
                    log_events.warning(f'Invalid json: {e_str}')
                    continue

                # Alive event from openhab to detect dropped connections
                # -> Can be ignored on the HABApp side
                if e_json.get('type') == 'ALIVE':
                    continue

                # Log raw sse event
                if log_events.isEnabledFor(logging.DEBUG):
                    log_events._log(logging.DEBUG, e_str, [])

                # process
                _see_handler(e_json)
    except Exception as e:
        disconnect = is_disconnect_exception(e)
        lvl = logging.WARNING if disconnect else logging.ERROR
        log.log(lvl, f'SSE request Error: {e}')
        for line in traceback.format_exc().splitlines():
            log.log(lvl, line)

        # reconnect even if we have an unexpected error
        if not disconnect:
            set_offline(f'Uncaught error in process_sse_events: {e}')


QUEUE = Queue()


async def output_queue_listener():
    # clear Queue
    try:
        while True:
            await QUEUE.get_nowait()
    except QueueEmpty:
        pass

    while True:
        try:
            while True:
                item, state, is_cmd = await QUEUE.get()

                if not isinstance(state, str):
                    state = convert_to_oh_type(state)

                if is_cmd:
                    await post(f'/rest/items/{item:s}', data=state)
                else:
                    await put(f'/rest/items/{item:s}/state', data=state)
        except Exception as e:
            process_exception(output_queue_listener, e, logger=log)


@ignore_exception
async def output_queue_check_size():

    first_msg_at = 150

    upper = first_msg_at
    lower = -1
    last_info_at = first_msg_at // 2

    while True:
        await sleep(5)
        size = QUEUE.qsize()

        # small log msg
        if size > upper:
            upper = size * 2
            lower = size // 2
            log_warning(log, f'{size} messages in queue')
        elif size < lower:
            upper = max(size / 2, first_msg_at)
            lower = size // 2
            if lower <= last_info_at:
                lower = -1
                log_info(log, 'queue OK')
            else:
                log_info(log, f'{size} messages in queue')


async def async_post_update(item, state: Any):
    QUEUE.put_nowait((item, state, False))


async def async_send_command(item, state: Any):
    QUEUE.put_nowait((item, state, True))


async def async_get_uuid() -> str:
    resp = await get('/rest/uuid', log_404=False)
    return await resp.text(encoding='utf-8')


async def async_get_root() -> dict:
    resp = await get('/rest/', log_404=False)
    if resp.status == 404:
        return {}
    return await resp.json(loads=load_json, encoding='utf-8')


async def async_get_system_info() -> dict:
    resp = await get('/rest/systeminfo', log_404=False)
    if resp.status == 404:
        return {}
    return await resp.json(loads=load_json, encoding='utf-8')


async def async_get_start_level(default_level: int = -10) -> int:
    system_info = await async_get_system_info()
    return system_info.get('systemInfo', {}).get('startLevel', default_level)


async def wait_for_min_start_level():

    waited_for_oh = False
    last_level = -100

    while True:
        start_level_is = await async_get_start_level()
        start_level_min = HABApp.CONFIG.openhab.general.min_start_level
        if start_level_is >= start_level_min:
            break

        # show msg only once
        if not waited_for_oh:
            log.info('Waiting for openHAB startup to be complete')

        # show start level change
        if last_level != start_level_is:
            log.debug(f'Startlevel: {start_level_is}')
        last_level = start_level_is

        # wait for openhab
        waited_for_oh = True
        await asyncio.sleep(1)

    # Startup complete
    if waited_for_oh:
        log.info('openHAB startup complete')


async def try_connect():
    global IS_ONLINE

    while True:
        try:
            # sleep before reconnect
            await CONNECT_WAIT.wait()

            log.debug('Trying to connect to OpenHAB ...')
            root = await async_get_root()

            # It's possible that we get status 4XX during startup and then the response is empty
            runtime_info = root.get('runtimeInfo')
            if not runtime_info:
                log.info('... offline!')
                continue

            log.info(f'Connected {"read only " if IS_READ_ONLY else ""}to OpenHAB '
                     f'version {runtime_info["version"]} ({runtime_info["buildString"]})')

            # todo: remove this 2023
            # Show warning (convenience)
            vers = tuple(map(int, runtime_info["version"].split('.')[:2]))
            if vers < (3, 3):
                log.warning('HABApp requires at least openHAB version 3.3!')

            # wait for openhab startup to be complete
            await wait_for_min_start_level()
            break
        except Exception as e:
            if isinstance(e, (OpenhabDisconnectedError, ExpectedSuccessFromOpenhab)):
                log.info('... offline!')
            else:
                for line in traceback.format_exc().splitlines():
                    log.error(line)

    IS_ONLINE = True

    # start sse processing
    TASK_SSE_LISTENER.start()

    # output messages
    TASK_QUEUE_WORKER.start()
    TASK_QUEUE_WATCHER.start()

    ON_CONNECTED()
    return None


TASK_SSE_LISTENER: Final = SingleTask(start_sse_event_listener, 'SSE event listener')
TASK_TRY_CONNECT: Final = SingleTask(try_connect, 'Try OH connect')

TASK_QUEUE_WORKER: Final = SingleTask(output_queue_listener, 'OhQueue')
TASK_QUEUE_WATCHER: Final = SingleTask(output_queue_check_size, 'OhQueueSize')


def __load_cfg():
    global IS_READ_ONLY
    IS_READ_ONLY = HABApp.config.CONFIG.openhab.general.listen_only


# setup config
__load_cfg()
HABApp.config.CONFIG.subscribe_for_changes(__load_cfg)


# import it here otherwise we get cyclic imports
from HABApp.openhab.connection_handler.sse_handler import on_sse_event  # noqa: E402
from HABApp.openhab.connection_handler.func_async import convert_to_oh_type  # noqa: E402



# no external dependencies
import HABApp.openhab.errors
import HABApp.openhab.events

# isort: split

import HABApp.openhab.interface_async
import HABApp.openhab.interface

# isort: split

# items use the interface for the convenience functions
import HABApp.openhab.items



import re
from pathlib import Path
from typing import Iterable, Optional, List, Dict

from HABApp.core.const.const import PYTHON_311
from HABApp.openhab.connection_logic.plugin_things.cfg_validator import UserItem
from .formatter import FormatterScope
from .formatter_builder import ValueFormatterBuilder, MultipleValueFormatterBuilder, ConstValueFormatterBuilder, \
    MetadataFormatter, LinkFormatter

if not PYTHON_311:
    from typing_extensions import Self
else:
    from typing import Self


FIELD_ORDER = (
    'type', 'name', 'label', 'icon', 'groups', 'tags', 'bracket_open', 'link', 'metadata', 'bracket_close'
)

RE_GROUP_NAMES = re.compile(r'([A-Za-z0-9-]+?)(?=[A-Z_ -])')


def brackets_needed(obj: UserItem):
    return obj.link or obj.metadata


class ItemsFileWriter:
    def __init__(self):
        self.items: List[UserItem] = []

    def add_item(self, obj) -> Self:
        self.items.append(obj)
        return self

    def add_items(self, objs: Iterable[UserItem]) -> Self:
        self.items.extend(objs)
        return self

    def group_items(self) -> List[List[UserItem]]:
        grouped_items: Dict[Optional[str], List[UserItem]] = {}
        not_grouped: List[UserItem] = []
        for item in self.items:
            if m := RE_GROUP_NAMES.match(item.name):
                grouped_items.setdefault(m.group(1), []).append(item)
            else:
                not_grouped.append(item)

        ret = []
        # sort alphabetical by key
        for key, values in sorted(grouped_items.items(), key=lambda x: x[0]):
            # if it's only one value it'll be not written in a block
            if len(values) <= 1:
                not_grouped.extend(values)
                continue

            ret.append(values)

        # single entry items get created last
        if not_grouped:
            ret.append(not_grouped)

        return ret

    def generate(self) -> str:
        groups = self.group_items()

        builder = {
            'type': ValueFormatterBuilder('type', '{:s}'),
            'name': ValueFormatterBuilder('name', '{:s}'),
            'label': ValueFormatterBuilder('label', '"{:s}"'),
            'icon': ValueFormatterBuilder('icon', '<{:s}>'),
            'groups': MultipleValueFormatterBuilder('groups', '{:s}', '({:s})'),
            'tags': MultipleValueFormatterBuilder('tags', '"{:s}"', '[{:s}]'),
            'bracket_open': ConstValueFormatterBuilder('{', condition=brackets_needed),
            'link': LinkFormatter(),
            'metadata': MetadataFormatter(),
            'bracket_close': ConstValueFormatterBuilder('}', condition=brackets_needed),
        }

        lines = []

        for group in groups:
            scope = FormatterScope(FIELD_ORDER, ('bracket_open', 'bracket_close', 'metadata'))
            for item in group:
                scope.add({k: v.create_formatter(item) for k, v in builder.items()})

            lines.extend(scope.get_lines())
            lines.extend([''])

        return '\n'.join(lines)

    def create_file(self, file: Path):

        output = self.generate()

        # only write changes
        if file.is_file():
            existing = file.read_text('utf-8')
            if existing == output:
                return False

        file.write_text(output, encoding='utf-8')
        return True



import logging
from typing import List, Tuple


class BufferedLogger:
    def __init__(self):
        self._msgs: List[Tuple[int, str]] = []

    def _log(self, lvl: int, msg: str):
        self._msgs.append((lvl, msg))

    def debug(self, msg: str) -> 'BufferedLogger':
        self._log(logging.DEBUG, msg)
        return self

    def info(self, msg: str) -> 'BufferedLogger':
        self._log(logging.INFO, msg)
        return self

    def warning(self, msg: str) -> 'BufferedLogger':
        self._log(logging.WARNING, msg)
        return self

    def error(self, msg: str) -> 'BufferedLogger':
        self._log(logging.ERROR, msg)
        return self

    def flush(self, logger: logging.Logger):
        for lvl, msg in self._msgs:
            logger.log(lvl, msg)
        self._msgs.clear()
